@@ -3,19 +3,32 @@ import dis
 import ctypes
 from collections import namedtuple
 from functools import partial
+from itertools import groupby
 from types import CodeType, FunctionType
 import logging
+import struct
 from importlib._bootstrap_external import _code_to_timestamp_pyc
 
 import subprocess
 import base64
 from shlex import quote
 import dill
+import sys
 
-from .mem_view import Mem, ptr_frame_stack_bottom, ptr_frame_stack_top
-from .minias import _dis, Bytecode, long2bytes
+from .mem_view import Mem
+from .py import (
+    JX,
+    ptr_frame_stack_bottom,
+    ptr_frame_stack_top,
+    ptr_frame_block_stack_bottom,
+    ptr_frame_block_stack_top,
+    put_NULL,
+    disassemble,
+)
+from .minias import _dis, long2bytes
 
 locals().update(dis.opmap)
+EXCEPT_HANDLER = 257
 
 
 def _overlapping(s1, l1, s2, l2):
@@ -119,7 +132,17 @@ def expand_long(c):
     return bytes(result)
 
 
-def get_value_stack_from_beacon(frame, beacon, expand=0):
+class NULL:
+    """Represents NULL"""
+    def __new__(cls):
+        return cls
+
+    def __str__(self):
+        return "<NULL>"
+    __repr__ = __str__
+
+
+def get_value_stack_from_beacon(frame, beacon, expand=0, null=NULL()):
     """
     Collects frame stack using beacon as
     an indicator of stack top.
@@ -131,20 +154,32 @@ def get_value_stack_from_beacon(frame, beacon, expand=0):
     beacon : int
         Value on top of the stack.
     expand : int
+    null
+        The NULL object replacement.
 
     Returns
     -------
     stack : list
         Stack contents.
     """
+    logging.debug(f"collecting stack for {frame} with beacon 0x{beacon:016x}")
     stack_bot = ptr_frame_stack_bottom(frame)
-    stack_view = Mem(stack_bot, (frame.f_code.co_stacksize + expand) * 8)[:]
+    stack_view = Mem(stack_bot, (frame.f_code.co_stacksize + expand) * 8)
+    logging.debug(f"  contents:\n{stack_view}")
+    stack_view = stack_view[:]
     result = []
     for i in range(0, len(stack_view), 8):
         obj_ref = int.from_bytes(stack_view[i:i + 8], "little")
+        logging.debug(f"  object at 0x{obj_ref:016x} ...")
         if obj_ref == beacon:
+            logging.debug(f"    <beacon>")
             return result
-        result.append(ctypes.cast(obj_ref, ctypes.py_object).value)
+        if obj_ref == 0:
+            result.append(null)
+            logging.debug(f"    (null)")
+        else:
+            result.append(ctypes.cast(obj_ref, ctypes.py_object).value)
+            logging.debug(f"    {repr(result[-1])}")
     raise RuntimeError("Failed to determine stack top")
 
 
@@ -172,13 +207,39 @@ def get_value_stack(frame):
     return result
 
 
-class FrameSnapshot(namedtuple("FrameSnapshot", ("code", "pos", "v_stack", "v_locals", "v_globals", "v_builtins"))):
+block_stack_item = namedtuple('block_stack_item', ('type', 'handler', 'level'))
+
+
+def get_block_stack(frame):
+    """
+    Collects block stack.
+
+    Parameters
+    ----------
+    frame : FrameObject
+        Frame to process.
+
+    Returns
+    -------
+    stack : list
+        Block stack contents.
+    """
+    fr = ptr_frame_block_stack_bottom(frame)
+    to = ptr_frame_block_stack_top(frame)
+    size = to - fr
+    size4 = size // 4
+    result = struct.unpack("i" * size4, Mem(fr, size)[:])
+    result = tuple(block_stack_item(*x) for x in zip(result[::3], result[1::3], result[2::3]))
+    return result
+
+
+class FrameSnapshot(namedtuple("FrameSnapshot", ("code", "pos", "v_stack", "v_locals", "v_globals", "v_builtins", "block_stack"))):
     """A snapshot of python frame"""
     slots = ()
     def __repr__(self):
         code = self.code
         contents = []
-        for i in "v_stack", "v_locals", "v_globals", "v_builtins":
+        for i in "v_stack", "v_locals", "v_globals", "v_builtins", "block_stack":
             v = getattr(self, i)
             if v is None:
                 contents.append(f"{i}: not set")
@@ -187,7 +248,7 @@ class FrameSnapshot(namedtuple("FrameSnapshot", ("code", "pos", "v_stack", "v_lo
         return f'FrameSnapshot {code.co_name} at "{code.co_filename}"+{code.co_firstlineno} @{self.pos:d} {" ".join(contents)}'
 
 
-def p_jump_to(pos, patcher, f_next):
+def p_jump_to(pos, patcher, f_next, jx=JX):
     """
     Patch: jump to position.
 
@@ -197,6 +258,8 @@ def p_jump_to(pos, patcher, f_next):
         Position to set.
     patcher : FramePatcher
     f_next : Callable
+    jx : int
+        Jump address multiplier.
 
     Returns
     -------
@@ -209,7 +272,7 @@ def p_jump_to(pos, patcher, f_next):
     else:
         logging.debug(f"jump_to {pos:d}: patching ...")
         if patcher.pos != pos - 2:
-            patcher.patch_current(expand_long([JUMP_ABSOLUTE, pos]), 2)  # jump to the original bytecode position
+            patcher.patch_current(expand_long([JUMP_ABSOLUTE, pos // jx]), 2)
         patcher.patch([CALL_FUNCTION, 0], pos)  # call next
         patcher.commit()
         logging.debug(f"jump_to {pos:d}: ⏎ {f_next}")
@@ -258,6 +321,8 @@ def p_place_beacon(beacon, patcher, f_next):
     -------
     f_next : Callable
         Next function to call.
+    beacon
+        The beacon object.
     """
     logging.debug(f"place_beacon {beacon}: patching ...")
     patcher.patch_current([
@@ -270,6 +335,31 @@ def p_place_beacon(beacon, patcher, f_next):
     return f_next, beacon
 
 
+def p_exit_block_stack(block_stack, patcher, f_next):
+    """
+    Patch: exits the block stack.
+
+    Parameters
+    ----------
+    block_stack
+        State of the block stack.
+    patcher : FramePatcher
+    f_next : Callable
+
+    Returns
+    -------
+    f_next : Callable
+        Next function to call.
+    """
+    logging.debug(f"exit block stack {len(block_stack):d} times")
+    patcher.patch_current(
+        [POP_BLOCK, 0] * len(block_stack) + [CALL_FUNCTION, 0], 2
+    )
+    patcher.commit()
+    logging.debug(f"exit block stack: ⏎ {f_next}")
+    return f_next
+
+
 def snapshot(frame, finalize, method="inject"):
     """
     Snapshot the stack starting from the given frame.
@@ -348,26 +438,42 @@ def snapshot(frame, finalize, method="inject"):
             assert prev_builtins is frame.f_builtins
 
         # save locals, globals, etc.
-        logging.info("  saving locals ...")
-        result.append(FrameSnapshot(
+        logging.info("  saving snapshot ...")
+        fs = FrameSnapshot(
             code=frame.f_code,
             pos=frame.f_lasti,
             v_stack=None if method == "inject" else get_value_stack(frame),
             v_locals=frame.f_locals.copy(),
             v_globals=prev_globals,
             v_builtins=prev_builtins,
-        ))
+            block_stack=get_block_stack(frame),
+        )
+        logging.info(f"    code: {fs.code}")
+        logging.info(f"    pos: {fs.pos}")
+        logging.info(f"    stack: {len(fs.v_stack) if fs.v_stack is not None else 'none'}")
+        logging.info(f"    locals: {len(fs.v_locals)}")
+        logging.info(f"    globals: {len(fs.v_globals)}")
+        logging.info(f"    builtins: {len(fs.v_builtins)}")
+        logging.info(f"    block_stack:")
+        if len(fs.block_stack):
+            for i in fs.block_stack:
+                logging.info(f"      {i}")
+        else:
+            logging.info("      (empty)")
+
+        result.append(fs)
 
         if method == "inject":  # prepare patchers
             logging.info(f"  patching the bytecode ...")
             original_code = bytearray(frame.f_code.co_code)  # store the original bytecode
             rtn_pos = original_code[::2].index(RETURN_VALUE) * 2  # figure out where it returns
-            # note that bytearray is intentional to guarantee the copy
+            # note that the bytearray is intentional to guarantee the copy
             patcher = FramePatcher(frame)
 
             p_jump_to(0, patcher, None)  # make room for patches immediately
             chain.append(partial(p_place_beacon, beacon, patcher))  # place the beacon
             chain.append(partial(notify, frame))  # collect value stack
+            chain.append(partial(p_exit_block_stack, fs.block_stack, patcher))  # exit from "finally" statements
             chain.append(partial(p_jump_to, rtn_pos - 2, patcher))  # jump 1 opcode before return
             chain.append(partial(
                 p_set_bytecode,
@@ -434,7 +540,7 @@ def pickle_generator(pickler, obj):
 
 
 def morph_execpoint(p, nxt, pack=None, unpack=None, globals=False, fake_return=True,
-        flags=0):
+                    flags=0):
     """
     Prepares a code object which morphs into the desired state
     and continues the execution afterwards.
@@ -466,9 +572,9 @@ def morph_execpoint(p, nxt, pack=None, unpack=None, globals=False, fake_return=T
         The resulting morph.
     """
     assert pack is None and unpack is None or pack is not None and unpack is not None,\
-        "Either both or none pack and unpack arguments should be specified"
+        "Either both or none pack and unpack arguments have be specified"
     logging.info(f"Preparing a morph into execpoint {p} pack={pack is not None} ...")
-    code = Bytecode.disassemble(p.code)
+    code = disassemble(p.code)
     code.pos = 0
     code.c("Header")
     code.nop(b'mrph')  # signature
@@ -510,10 +616,27 @@ def morph_execpoint(p, nxt, pack=None, unpack=None, globals=False, fake_return=T
 
     # stack
     if len(p.v_stack) > 0:
-        code.c(f"*stack")
-        v_stack = p.v_stack[::-1]
-        _LOAD(v_stack)
-        code.i(UNPACK_SEQUENCE, len(v_stack))
+        for k, v_stack in groupby(p.v_stack, lambda x: x is not NULL):
+            if k:
+                code.c("*stack")
+                v_stack = tuple(v_stack)[::-1]
+                _LOAD(v_stack)
+                code.i(UNPACK_SEQUENCE, len(v_stack))
+            else:
+                code.c("*NULL")
+                for _ in v_stack:
+                    put_NULL(code)
+
+    # block stack
+    if len(p.block_stack) > 0:
+        code.c("block_stack")
+        for i, (type, handler, level) in enumerate(p.block_stack):
+            if type == SETUP_FINALLY:
+                code.i(SETUP_FINALLY, 0, jump_to=code.by_pos(handler * JX))
+            elif type == EXCEPT_HANDLER:
+                raise NotImplementedError(f"'except:' not implemented")
+            else:
+                raise NotImplementedError(f"Unknown block type={type} ({dis.opname.get(type, 'unknown opcode')})")
 
     if nxt is not None:
         # call nxt which is a code object
@@ -529,16 +652,8 @@ def morph_execpoint(p, nxt, pack=None, unpack=None, globals=False, fake_return=T
         code.I(LOAD_CONST, None)  # fake nxt returning None
 
     # now jump to the previously saved position
-    target_pos = p.pos + 2  # p.pos points to the last executed opcode
-    # find the instruction ...
-    for jump_target in code.iter_opcodes():
-        if jump_target.pos == target_pos:
-            break
-    else:
-        raise RuntimeError
-    # ... and jump to it (the argument will be determined after re-assemblling the bytecode)
     code.c(f"goto saved pos")
-    code.i(JUMP_ABSOLUTE, 0, jump_to=jump_target)
+    code.i(JUMP_ABSOLUTE, 0, jump_to=code.by_pos(p.pos + 2))
 
     code.c(f"---------------------")
     code.c(f"The original bytecode")
@@ -625,7 +740,7 @@ def bash_inline_create_file(name, contents):
     return f"echo {quote(base64.b64encode(contents).decode())} | base64 -d > {quote(name)}"
 
 
-def shell_teleport(*shell_args, python="python", before="cd $(mktemp -d)",
+def shell_teleport(*shell_args, python=None, before="cd $(mktemp -d)",
         pyc_fn="payload.pyc", shell_delimeter="; ", pack_file=bash_inline_create_file,
         pack_object=dill.dumps, unpack_object=("dill", "loads"),
         _frame=None, **kwargs):
@@ -662,6 +777,8 @@ def shell_teleport(*shell_args, python="python", before="cd $(mktemp -d)",
     -------
     None
     """
+    if python is None:
+        python = sys.executable
     payload = []
     if not isinstance(before, (list, tuple)):
         payload.append(before)
